((nil . ((indent-tabs-mode . nil)
         (fill-column . 100)))
 (c-mode . ((c-file-style . "gnu")
            (c-file-offsets
             (brace-list-intro . +))))
 (c-ts-mode . ((c-ts-mode-style . gnu)))
)
